from pathlib import Path
import json
from typing import Dict
from jinja2 import Template
import black
import argparse
from collections import defaultdict
from datetime import datetime

BASE_DIR = Path(__file__).parent.parent / "types"
OUTPUT_FILE = "central_schema.py"
EXCLUDED_SCHEMAS = {"Item", "Type", "PluginRun"}
SCHEMA_TYPES = {
    "Bool": bool,
    "Text": str,
    "Integer": int,
    "Real": float,
    "DateTime": datetime
}

HEADER = """# AUTOGENERATED, DO NOT EDIT!
# This file was generated by /tools/generate_central_schema.sh
# Visit https://gitlab.memri.io/memri/schema to learn more 

from datetime import datetime
from typing import Optional, Any

from pymemri.data.itembase import Item, EdgeList
"""

TEMPLATE = '''
class {{name}}({{base_schema}}):
    description = """{{description}}"""
    {% if properties is defined %}
    properties = {{base_schema}}.properties + {{properties|map(attribute="name")|list}}{% endif %}    
    {% if edges is defined %}
    edges = {{base_schema}}.edges + {{edges|map(attribute="name")|list}}{% endif %}
    
    
    def __init__(
        self,
        {% for property in properties %}
        {{property.name}}: {{property.type}} = None,
        {% endfor %}
        {% for edge in edges %}
        {{edge.name}}: EdgeList["{{edge.target}}"] = None,
        {% endfor %}
        **kwargs
    ):
        super().__init__(**kwargs)

        {% if properties is defined and properties|length > 0%}
        # Properties
        {% for property in properties %}
        self.{{property.name}}: Optional[{{property.type}}] = {{property.name}} 
        {% endfor %}{% endif %}

        {% if edges is defined and edges|length > 0%}
        # Edges
        {% for edge in edges %}
        self.{{edge.name}}: EdgeList["{{edge.target}}"] = EdgeList("{{edge.name}}", "{{edge.target}}", {{edge.name}})
        {% endfor %}{% endif %}
'''


def generate_from_json(path, schema_template, verbose=True):
    name = path.stem

    with open(path, "r") as f:
        item_dict = json.load(f)

    base_cls = item_dict.get("base_schema", "Item")
    properties = []
    for property in item_dict["properties"]:
        if property["type"] in SCHEMA_TYPES:
            property["type"] = SCHEMA_TYPES[property["type"]].__name__
            properties.append(property)
        else:
            raise ValueError(f"Unknown property type '{property['type']}' in {str(path)}")
    
    py_str = schema_template.render(
        name=name,
        base_schema=base_cls,
        properties=properties,
        edges=item_dict["edges"],
        description=item_dict["description"],
    )
    return name, py_str, base_cls


def get_schema_order(inheritance: defaultdict):
    # Remove default entry to pop items
    inheritance = dict(inheritance)
    schema_order = sorted(inheritance.pop("Item"))

    # Schema order:
    # - Superclasses first, subclasses after.
    # - Sorted alphabetically
    while len(inheritance):
        remaining_len = len(inheritance)
        for base_cls in sorted(inheritance.keys()):
            if base_cls in schema_order:
                schema_names = inheritance.pop(base_cls)
                schema_order.extend(sorted(schema_names))
        if remaining_len == len(inheritance):
            raise ValueError(f"Could not resolve inheritance. Missing base types: {inheritance.keys()}")

    return schema_order

parser = argparse.ArgumentParser()
parser.add_argument("-o", "--output_file", metavar="", type=str, default="./central_schema.py",
                    help="Path of the output file, default './central_schema.py'")
parser.add_argument("-v", "--verbose", default=False, action="store_true")

if __name__=="__main__":
    args = parser.parse_args()

    schema_template = Template(TEMPLATE, lstrip_blocks=True, trim_blocks=True)
    inheritance = defaultdict(lambda: list())    
    generated_schema = dict() # dict[name, schem_str]
    for path in BASE_DIR.rglob("*.json"):
        if path.stem in EXCLUDED_SCHEMAS:
            continue
        schema_name, schema_str, base_cls = generate_from_json(path, schema_template, args.verbose)
        generated_schema[schema_name] = schema_str
        inheritance[base_cls].append(schema_name)

    schema_order = get_schema_order(inheritance)
    schema_ordered = [generated_schema[name] for name in schema_order]
    file_content = "\n\n".join(schema_ordered)
    file_content = HEADER + "\n\n" + file_content
    file_content = black.format_str(file_content, mode=black.FileMode())

    print(f"writing to {args.output_file}")
    with open(args.output_file, 'w') as f:
        f.write(file_content)
